<?php

namespace Drupal\fieldupdate\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;

/**
 * Class ForgotUsernameForm.
 */
class FieldUpdateAdvance extends FormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'fieldupdateadvance';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $entity_types_options = [
      'node' => $this->t('Content'),
      'taxonomy_term' => $this->t('Taxonomy term'),
      'user' => $this->t('User'),
    ];
    $trigger_element = $form_state->getUserInput()['_triggering_element_name'] ?? NULL;
    $selected_entity = $form_state->getValue('entity_types');
    if ('entity_types' == $trigger_element) {
      $selected_bundle = $selected_field = '';
    }
    if ('bundle' == $trigger_element) {
      $selected_bundle = $form_state->getValue('bundle');
      $selected_field = '';
    }
    elseif ('fields' == $trigger_element || empty($trigger_element)) {
      $selected_bundle = $form_state->getValue('bundle');
      $selected_field = $form_state->getValue('fields');
    }

    $form['info'] = [
      '#markup' => $this->t('<p>Update entities based on query or entity id.</p>',),
    ];
    $form['entity_types'] = [
      '#title' => t('Entity type'),
      '#type' => 'select',
      '#options' => $entity_types_options,
      '#empty_option' => $this->t('- None -'),
      '#ajax' => [
        'callback' => '::updateBundleDropdownOptions',
        'wrapper' => 'bundle-field-wrapper',
        'event' => 'change',
      ],
      '#default_value' => $selected_entity,
      '#required' => TRUE,
    ];
    $form['depended_fieldset_container'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'bundle-field-wrapper'],
    ];

    $form['depended_fieldset_container']['bundle'] = [
      '#type' => 'select',
      '#title' => $this->t('Bundle'),
      '#options' => $this->getBundleDropdownOptions($selected_entity),
      '#default_value' => $selected_bundle,
      '#empty_option' => $this->t('- None -'),
      '#ajax' => [
        'callback' => '::fieldsDropdownCallback',
        'wrapper' => 'fields-field-wrapper',
      ],
      '#states' => [
        'required' => [
          ':input[name="entity_types"]' => ['value' => 'node'],
        ],
      ],
    ];
    $form['depended_fieldset_container']['depend_next'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'fields-field-wrapper'],
    ];
    $form['depended_fieldset_container']['depend_next']['fields'] = [
      '#type' => 'select',
      '#title' => $this->t('Field List') . $form_state->getValue('fields'),
      '#options' => static::getFiledsForSelections($selected_entity, $selected_bundle),
      '#default_value' => $selected_field,
      '#required' => TRUE,
      '#empty_option' => $this->t('- None -'),
      '#ajax' => [
        'callback' => '::fieldsForUpdate',
        'wrapper' => 'entity-field-wrapper',
      ],

    ];
    $form['depended_fieldset_container']['depend_next']['entity_field_wrapper'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'entity-field-wrapper'],
    ];
    if ($selected_entity) {
      $selected_bundle = $selected_bundle ?: $selected_entity;
      if ($selected_field) {
        $form['depended_fieldset_container']['depend_next']['entity_field_wrapper'] = $this->renderFormFieldForUpdate($selected_entity, $selected_bundle, $selected_field);
      }
      else {
        $form['depended_fieldset_container']['depend_next']['entity_field_wrapper']['update_field'] = [
          '#type' => 'container',
          '#attributes' => ['id' => 'entity-field-wrapper'],
        ];
      }

    }
    $form['submit'] = [
      '#value' => t('Update'),
      '#type' => 'submit',
    ];
    $form['cancel'] = [
      '#markup' => '<a class="button cancel" href="' . Url::fromRoute('fieldupdate.advanceupdate')->toString() . '">Reset</a>',
    ];
    return $form;
  }

  /**
   * On Ajax get options for entity type dropdown.
   *
   * @param array $form
   *   Accept form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Accept FormStateInterface.
   *
   * @return mixed
   *   Return form rendered array.
   */
  public function updateBundleDropdownOptions(array &$form, FormStateInterface $form_state) {
    return $form['depended_fieldset_container'];
  }

  /**
   * On Ajax get options for bundle dropdown.
   *
   * @param array $form
   *   Accept form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Accept FormStateInterface.
   *
   * @return mixed
   *   Return form rendered array.
   */
  public function fieldsDropdownCallback(array &$form, FormStateInterface $form_state) {
    return $form['depended_fieldset_container']['depend_next'];

  }

  /**
   * On Ajax get options for field dropdown.
   *
   * @param array $form
   *   Accept form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Accept FormStateInterface.
   *
   * @return mixed
   *   Return form rendered array.
   */
  public function fieldsForUpdate(array &$form, FormStateInterface $form_state) {
    return $form['depended_fieldset_container']['depend_next']['entity_field_wrapper'];
  }

  /**
   * Get options for field dropdown.
   *
   * @param string $entity
   *   Accept entity type string.
   *
   * @return array
   *   Return form rendered array.
   */
  public function getBundleDropdownOptions($entity = '') {
    $types = [];
    $entity_machine_name = [
      'node' => 'node_type',
      'taxonomy_term' => 'taxonomy_vocabulary',
    ];
    if ($entity && array_key_exists($entity, $entity_machine_name)) {
      $entityTypeManager = \Drupal::service('entity_type.manager');
      $contentTypes = $entityTypeManager->getStorage($entity_machine_name[$entity])
        ->loadMultiple();
      foreach ($contentTypes as $contentType) {
        $types[$contentType->id()] = $contentType->label();
      }
    }
    return $types;
  }

  /**
   * Get options for field dropdown.
   *
   * @param string $selected_entity
   *   Accept entity type string.
   * @param string $bundle
   *   Accept bundle type string.
   *
   * @return array
   *   Return form rendered array.
   */
  private function getFiledsForSelections($selected_entity, $bundle = '') : array {
    $types = [];
    if ('user' == $selected_entity) {
      $bundle = $selected_entity;
    }
    if ($selected_entity && $bundle) {
      $entityFieldManager = \Drupal::service('entity_field.manager');
      $fields = $entityFieldManager->getFieldDefinitions($selected_entity, $bundle);
      foreach ($fields as $key => $field) {
        $info = FieldStorageConfig::loadByName($selected_entity, $key);
        if ($info) {
          $types[$field->getName()] = $field->getLabel() . " (" . $field->getName() . ")";
        }
      }

    }
    return $types;
  }

  /**
   * Generate form elements to updated on main form.
   *
   * @param string $selected_entity
   *   Accept entity type string.
   * @param string $selected_bundle
   *   Accept bundle type string.
   * @param string $selected_field
   *   Accept field machine name.
   *
   * @return array
   *   Return form rendered array.
   */
  private function renderFormFieldForUpdate(string $selected_entity, string $selected_bundle, string $selected_field): array {

    $field_config = FieldConfig::loadByName($selected_entity, $selected_bundle, $selected_field);
    $storage_config = FieldStorageConfig::loadByName($selected_entity, $selected_field);

    $update_count = $this->getEntityForUpdate($selected_entity, $selected_bundle, $selected_field);
    $return = [];
    $field_label = $field_config->getLabel();
    $field_name = $field_config->getName();
    $field_type = $field_config->getType();

    $drupal_fields = [
      'image' => 'textfield',
      'string' => 'textfield',
      'integer' => 'textfield',
      'boolean' => 'checkbox',
      'text_long' => 'text_format',
      'text_with_summary' => 'text_format',
      'entity_reference' => 'entity_autocomplete',
      'list_string' => 'select',
      'list_integer' => 'select',
      'datetime' => 'datetime',
      'timestamp' => 'datetime',
    ];
    if (array_key_exists($field_type, $drupal_fields)) {
      $form_element = [];
      switch ($field_type) {
        case 'string':
        case 'integer':
          $form_element['#type'] = 'textfield';
          break;

        case 'image':
          $form_element['#type'] = 'textfield';
          $form_element['#description'] = 'Add full file path. Like https://example.com/image.png';
          break;

        case 'boolean':
          $form_element['#type'] = 'checkbox';
          break;

        case 'text_long':
        case 'text_with_summary':
          $form_element['#type'] = 'text_format';
          break;

        case 'list_string':
        case 'list_integer':
          $allowed_values = $storage_config->getSettings()['allowed_values'];
          $form_element['#type'] = 'select';
          $form_element['#options'] = $allowed_values;
          break;

        case 'entity_reference':
          $form_element['#type'] = 'entity_autocomplete';
          $form_element['#target_type'] = $storage_config->getSettings()['target_type'];
          $form_element['#selection_settings'] = ['target_bundles' => $field_config->getSettings()['handler_settings']['target_bundles']];
          break;

        case 'datetime':
        case 'timestamp':
          $form_element['#type'] = 'datetime';
          break;
      }
      $form_element['#title'] = $field_label;
      $form_element['#attributes'] = ['id' => 'entity-field-wrapper'];
      $form_element['#required'] = TRUE;

      $append['el_container'] = [
        '#type' => 'container',
        '#attributes' => ['id' => 'entity-field-wrapper'],
      ];
      $append['el_container']['update_field'] = $form_element;
      $append['el_container']['ids'] = [
        '#type' => 'textarea',
        '#title' => 'Entity Ids',
        '#attributes' => ['id' => 'entity-field-wrapper'],
        '#required' => TRUE,
        '#description' => 'Add entity id seperated by commas. The given ids are updated with new values.',
      ];
      $append['el_container']['update_fieldtype'] = [
        '#type' => 'hidden',
        '#value' => $field_config->getType(),
      ];

      return $append;
    }

    return [
      '#markup' => t('@type type update not support for now', ['@type' => $field_type]),
      'id' => 'entity-field-wrapper',
    ];

  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    if ($form_state->getValue('op')) {

    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $post = $form_state->getValues();
    $entity_type = $post['entity_types'];
    $bundle = $post['bundle'];
    $field_machine_name = $post['fields'];
    $field_data = $post['update_field'];
    $field_type = $post['update_fieldtype'];

    if ('image' == $field_type) {
      $data = file_get_contents($field_data);
      if ($data) {
        $field_config = FieldConfig::loadByName($entity_type, $bundle, $field_machine_name);
        $field_settings = $field_config->getSettings();
        $token = \Drupal::token();
        $upload_location = $field_settings['uri_scheme'] . '://' . $token->replace($field_settings['file_directory']) . '/' . basename($field_data);
        $file = \Drupal::service('file.repository')->writeData($data, $upload_location);
        $field_data = [
          'target_id' => $file->id(),
          'alt' => '',
          'title' => '',
        ];
      }
    }
    elseif ('datetime' == $field_type) {
      $field_data = date('Y-m-d\TH:i:s', $field_data->getTimestamp());
    }
    elseif ('timestamp' == $field_type) {
      $field_data = $field_data->getTimestamp();
    }

    $results = array_map('trim', preg_split("/,/", $post['ids']));
    $total = count($results);

    $batch = [
      'title' => t('Importing @entity', ['@entity' => $entity_type]),
      'operations' => [],
      'init_message' => t('Import process is starting.'),
      'progress_message' => t('Processed @current out of @total. Estimated time: @estimate.'),
      'error_message' => t('The process has encountered an error.'),
    ];

    foreach ($results as $entity_id) {
      $dt = [
        'entity_id' => $entity_id,
        'entity_type' => $entity_type,
        'value' => $field_data,
        'field_machine_name' => $field_machine_name,
      ];
      $batch['operations'][] = [
        ['\Drupal\fieldupdate\Form\FieldUpdateAdvance', 'fieldupdateOperations'],
        [$dt],
      ];
    }
    batch_set($batch);
    \Drupal::messenger()
      ->addMessage(t('Updated @total entities!', ['@total' => $total]));

    $form_state->setRebuild(TRUE);

  }

  /**
   * Perform bulk operation to update the field value.
   *
   * @param array $data
   *   Data for batch operation.
   * @param mixed $context
   *   Content reference variable for bulk operation.
   *
   * @return void
   *   No return.
   */
  public static function fieldupdateOperations($data, &$context): void {
    $entity = \Drupal::entityTypeManager()
      ->getStorage($data['entity_type'])
      ->load($data['entity_id']);
    if ($entity) {
      $entity->set($data['field_machine_name'], $data['value']);
      $entity->save();
      $context['results'][] = $entity->label();
      $context['message'] = t('Created @title', ['@title' => $entity->label()]);
    }
  }

  /**
   * Get number of updated entity with the operation.
   *
   * @param string $entity_type
   *   Accept entity type string.
   * @param string $bundle
   *   Accept bundle type string.
   * @param string $field_machine_name
   *   Accept field machine name.
   *
   * @return int
   *   Return entity count for bulk action.
   */
  private function getEntityForUpdate(string $entity_type, string $bundle, string $field_machine_name): int {
    $query = \Drupal::entityQuery($entity_type);
    $query->accessCheck(TRUE);
    if ('node' == $entity_type) {
      $query->condition('type', $bundle);
    }
    if ('taxonomy_term' == $entity_type) {
      $query->condition('vid', $bundle);
    }
    if ('user' == $entity_type) {
      $query->condition('uid', 0, '>');
    }
    $query->notExists($field_machine_name);
    return $query->count()->execute();
  }

}
